/* Elements of matrix algebra and MULTIPLE REGRESSION
EXAMPLE how to use regression:
main () {
  float x[4][2] = {{1,0},{2,1},{3,0},{4,1}};
  float y[4] = {0,1,2,3.1};
  float b[2], b0, cov[2][2], rsq;
  if(matrRegress(x[0], y, b, &b0, cov[0], &rsq, 2, 4)==1)
     printf("ERROR");
}
*/
#include "matrix.h"
#include "mylib.h"

#define  ELEM(b,x,y)  b[x*n*2+y]

static float *point_weights;
static float ss_y, ss_z, ss_x; /* Sums of squares: total, residual, factorial */

static double *dblCov, *dblXY;
static float *x_aver, y_aver, sum_wgt;

/*******************************************/
int   matrRegress  (float *x, float *y, float *b,
			 float *b0, float *cov, float *rsq, int nfac, int n)
/*******************************************
DESCRIPTION: Regression analysis.
n - number of replications
nfac - number of factors
x[n*nfac] - plan of experiment (x1, x2,...)
y[n]  - responce
b[nfac] - regression coefficients
*b0 - intercept  (passed by pointer)
cov[nfac*nfac] - covariance matrix multiplied by n.
*rsq - R-square (passed by pointer)
RETURNS: 0-success, 1-error (factors lineary dependent)
*/
{
float *rev,    /* reverted covariance matrix */
	*norm,      /* centered matrix 'x' */
	*trans,     /* transposed matrix 'norm' */
	*yx,        /* sum of x*y products */
	*ypred,     /* predicted y */
	*y_wgt;     /* weighted y */
int i, j, ret, def_weights;
float dif;

	rev   = matrCreate(nfac*nfac);
	norm  = matrCreate(n*nfac);
	trans = matrCreate(n*nfac);
	yx    = matrCreate(nfac);
	ypred = matrCreate(n);
	y_wgt = matrCreate(n);
	x_aver  = matrCreate(nfac);

	if (point_weights == NULL){
		def_weights = FALSE;
		point_weights = matrCreate(n);
		for (i=0; i<n; ++i)
			point_weights[i] = 1;
		}
	else
		def_weights = TRUE;

   matrCopy(norm, x, n*nfac);
   matrAver(norm, x_aver, n, nfac);
   for (j=0; j<nfac; ++j)
      for (i=0; i<n; ++i)
			norm[i*nfac+j] -= x_aver[j];
	matrTrans(norm, trans, n, nfac);
	matrWeight(trans, point_weights, trans, n, nfac);

	matrMult(trans, norm, cov, nfac, n, nfac);

   matrMult(trans, y, yx, nfac, n, 1);
   matrCopy(rev, cov, nfac*nfac);
   ret = matrReverse(rev, nfac);
   matrMult(rev, yx, b, nfac, nfac, 1);

   matrAver(y, &y_aver, n, 1);
   matrMult(x_aver, b, b0, 1, nfac, 1);
	*b0 = y_aver - *b0;

	matrWeight(y, point_weights, y_wgt, n, 1);
	matrMult(y, y_wgt, &ss_y, 1, n, 1);
	sum_wgt = 0;
	for (i=0; i<n; ++i)
		sum_wgt += point_weights[i];
	ss_y -= y_aver * y_aver * sum_wgt;
   matrMult(norm, b, ypred, n, nfac, 1);
	ss_z = 0;
   for (i=0; i<n; ++i){
      ypred[i] += y_aver;
      dif = ypred[i] - y[i];
		ss_z += dif*dif*point_weights[i];
      }
	ss_x = ss_y - ss_z;
	if (ss_y > 0.0)
		*rsq = ss_x / ss_y;
	else
		*rsq = 0;

	if (def_weights == FALSE){
		free(point_weights);
		point_weights = NULL;
		}
   free (rev);
	free (norm);
	free (trans);
	free (yx);
	free (y_wgt);
   free (ypred);
   free (x_aver);
   return(ret);
}

/*************************************/
void accumRegress (float *x, float y, float wgt,
			int number, int nfac)
/*************************************/
{
	int i,j;

	if (number == 0)  /* start of accumulation */
	{
		check(dblCov = (double*)calloc(nfac*nfac,sizeof(double)));
		ss_y = 0;
		x_aver = matrCreate(nfac);
		y_aver = 0;
		check(dblXY = (double*)calloc(nfac,sizeof(double)));
		sum_wgt = 0;
	}
	for (i=0; i<nfac; ++i)
	{
		for (j=0; j<nfac; ++j)
			dblCov[i*nfac+j] += (double)x[i]*x[j]*wgt;
		x_aver[i] += x[i]*wgt;
		dblXY[i] += (double)x[i]*y*wgt;
	}
	ss_y += y*y*wgt;
	y_aver += y*wgt;
	sum_wgt += wgt;
}

/*************************************/
int  estimateRegress (float *b, float *b0, int nfac)
/*************************************/
{
	int i, j, ret;
	float *cov, *xy, *rev;

	cov = matrCreate(nfac*nfac);
	rev = matrCreate(nfac*nfac);
	xy = matrCreate(nfac);

	y_aver /= sum_wgt;
	for (i=0; i<nfac; ++i)
		x_aver[i] /= sum_wgt;
	for (i=0; i<nfac; ++i)
	{
		for (j=0; j<nfac; ++j)
			cov[i*nfac+j] = dblCov[i*nfac+j] -
			 (double)x_aver[i]*x_aver[j]*sum_wgt;
		xy[i] = dblXY[i] - (double)x_aver[i]*y_aver*sum_wgt;
	}
	ss_y -= y_aver*y_aver*sum_wgt;

	matrCopy(rev, cov, nfac*nfac);
	ret = matrReverse(rev, nfac);
	matrMult(rev, xy, b, nfac, nfac, 1);

	matrMult(x_aver, b, b0, 1, nfac, 1);
	*b0 = y_aver - *b0;

	free (rev);
	free (cov);
	free (dblCov);
	free (xy);
	free (dblXY);
	free (x_aver);
	return(ret);
}

/*************************************/
void  swap_rows (int *ind, int i, int j)
/*************************************/
{
	int l;
	l = ind[i];
	ind[i] = ind[j];
	ind[j] = l;
}

/*************************************/
int  matrReverse  (float *a, int n)
/*************************************
a[n*n],b[n*2n], ind[n]
a[n*n] - input and output
ind - number of row after switching
RETURNS: 0 on success, 1 on error.
*/
{
float *b;
int *ind;
int i, j, k, l, swaprows, found;
float c;

   b = (float*)calloc(2*n*n,sizeof(float));
	ind = (int*)calloc(n,sizeof(int));
  /* unity matrix + input */

	for (i = 0; i < n; ++i){
		for (j = 0; j < n; ++j){
			ELEM(b, i, j + n) = 0;
			ELEM(b, i, j) = a[i*n+j];
			}
		ELEM(b, i, i + n) = 1;
		}
	for (i = 0; i < n; ++i)
		ind[i] = i;

  /* Swapping the rows */

	do{
		swaprows = 0;
		i = 0;
		while (i < n && ELEM(b, ind[i], i) != 0)
			++i;
		if (i < n){
			swaprows = 1;
			for (; i < n; ++i){
				if (ELEM(b, ind[i], i) == 0){
					j = i + 1;
					while (j < n && ELEM(b, ind[j], i) == 0)
						++j;
					if (j < n){
						swap_rows (ind, i, j);
						continue;
						}
					j = 0;
					do{
						while (j < i && ELEM(b, ind[j], i) == 0)
							++j;
						if (j == i)
						{
							free(b);
							free(ind);
							return (1);
						}
						if (ELEM(b, ind[i], j) != 0){
							swap_rows (ind, i, j);
							break;
							}
						found = FALSE;
						for (k = i + 1; k < n && !found; ++k){
							if (ELEM(b, ind[k], j) != 0){
								found = TRUE;
								swap_rows (ind, k, j);
								swap_rows (ind, i, j);
								}
							}
						if (found)
							break;
						}while (TRUE);
					}
				}
			}
		} while (swaprows);

	for (i = 0; i < n; ++i){
		c = ELEM(b, ind[i], i);
		if (c == 0.0)
		{
			free(b);
			free(ind);
			return(1);
		}
		for (j = 0; j < n*2; ++j)
			ELEM(b, ind[i], j) = ELEM(b, ind[i], j) / c;
		for (k = i+1; k < n; ++k){
			c = ELEM(b, ind[k], i);
			if (c != 0)
			for (j = 0; j < n*2; ++j)
				ELEM(b,ind[k],j) = ELEM(b,ind[k],j) / c - ELEM(b,ind[i],j);
			}
		}
   /* Matrix is triangle now do the same back */

	for (i = n-1; i >=0; --i){
		for (k = 0; k < i; ++k){
			c = ELEM(b, ind[k], i);
			for (j = 0; j < n*2; ++j)
				ELEM(b,ind[k],j) = ELEM(b,ind[k],j) - c * ELEM(b, ind[i], j);
			}
		}
   for (i = 0; i < n; ++i)
		for (j = 0; j < n; ++j)
			a[i*n+j] = ELEM(b, ind[i], j + n);
   free(b);
   free(ind);
	return(0);
}

/********************************/
void  matrMult  (float *a, float *b, float *c, int n, int m, int k)
/********************************
a[n*m] X b[m*k] = c[n*k]
*/
{
int i, j, l;
   for (i=0; i<n; ++i){
      for (j=0; j<k; ++j){
			c[i*k+j] = 0;
			for (l=0; l<m; ++l)
				c[i*k+j] += a[i*m+l]*b[l*k+j];
		}
   }
}
/******************************/
float  *matrCreate  (int n)
/******************************/
{
float *pt;
	check (pt = (float*)calloc(n, sizeof(float)));
	return(pt);
}
/*************************************************/
void   matrCopy   (float *a, float *b, int n)
/*************************************************/
{
int i;
   for (i=0; i<n; ++i)
      a[i] = b[i];
}

/**************************************************/
void   matrTrans   (float *a, float *b, int n, int m)
/**************************************************/
{
int i, j;
   for (i=0; i<n; ++i)
		for (j=0; j<m; ++j)
	 b[j*n+i] = a[i*m+j];
}
/*******************************************/
void  matrAver (float *a, float *aver, int n, int ncol)
/*******************************************/
{
int i,j;
float sum_wgt=0;

	if (point_weights != NULL)
		for (i=0; i<n; ++i)
			sum_wgt += point_weights[i];
	else
		sum_wgt = n;
   for (j=0; j<ncol; ++j){
      aver[j] = 0;
		for (i=0; i<n; ++i){
			if (point_weights != NULL)
				aver[j] += a[i*ncol+j] * point_weights[i];
			else
				aver[j] += a[i*ncol+j];
			}
		aver[j] /= sum_wgt;
   }
}

/*******************************************/
void   create_point_weights  (float *wght, int n)
/*******************************************/
{
	point_weights = matrCreate(n);
	matrCopy(point_weights, wght, n);
}

/*******************************************/
void   destroy_point_weights   (void)
/*******************************************/
{
	free(point_weights);
	point_weights = NULL;
}

/*******************************************/
void   matrWeight  (float *a, float *wgt, float *b, int n, int repl)
/*******************************************/
{
int i, j;

	for (i=0; i<repl; ++i)
		for (j=0; j<n; ++j)
		b[i*n+j] = a[i*n+j]*wgt[j];
}
/*******************************************/
void  matrAddConst  (float *a, float c, int n)
/*******************************************/
{
int i;

	for (i=0; i<n; ++i)
		a[i] += c;
}
/*******************************************/
void  get_sum_squares  (float *sx, float *sy, float *sz)
/*******************************************/
{
	*sx = ss_x;
	*sy = ss_y;
	*sz = ss_z;
}